#! /usr/bin/perl
#
# Copyright (c) 2002 The Regents of the Univ. of California. All rights reserved.
#
# Permission is hereby granted, without written agreement and without license
# or royalty fees, to use, copy, modify, and distribute this software and its
# documentation for any purpose, provided that the above copyright notice and
# the following two paragraphs appear in all copies of this software.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN
# "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE
# MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.]
#
# Synopsis:    PROCESSING DC FILE
#
# Description: This file implements a translater from the declarative code (DC),
#              result from an Esterel compiler, into an extended BLIFMV file.
#
# Limitations: This converter only works on the subset of DC formats generated
#              by Esterel compiler with following limitations:
#  -- no [package|action|instance] table
#  -- no structured types (array|record)
#  -- no [equ|memo] cases, no [node|action] calls
#  -- no window 
#
# Author:      Yunjian William Jiang
# Date:        04/22/2002
# History:     
#     04/29    first version
#
#
# TODOs:
#  -- produce variable type info in generated header files (function/procedure)
#  -- procedure calls depends on dynamic variable value; 
#     may be treated by adding an additional mux node
#       74: equ: 120 48 at: 119; 
#       75: proccall: 14()(120) at: 118; 
#
#
# --------------Data structure:--------------
# $Parse::pi     \[$flow]
# $Parse::po     \[$flow]
# @Parse::nets    [$flow]
# @Parse::nodes   [""]
# @Parse::const   [$const]
# @Parse::equs    [$equ]
# @Parse::mems    [$equ]
# 
# $type  "type_name" | "integer.type_name"
# $const \{NAME=>"", TYPE=>$type, VALUE=>""}
# $flow  \{NAME=>"", TYPE=>$type, DEFAULT=>"", ASSERT=>""}
# $func  \{NAME=>"", PARAM=>[$param], TYPE=>$type, PROP=>""}
# $proc  \{NAME=>"", PARAMIN=>[$param], PARAMOUT=>[$param], PROP=>""}
# $param \{NAME=>"", TYPE=>$type,}
# $equ   \{EXPR=>$expr, COND=>$cond,}
# $expr  \{TYPE=>"CONST|FLOW|FUNC", CONST=>"", FLOW=>"", FUNC=>$impl, }
# $impl  \{NAME=>"", LIST=>[$expr],}
#

use strict;
use DCParse;


#--------------------------------------------------------------------#
my $file;
my ($verbose,$test_dc,$trace_on,$print_blifmv,$print_head) = (0,0,0,0,0);
my ($header_file);
my $Usage="perl $0 [-v n] [-h] [-c] [-t] [-d] [-p] <dc_file.dc>
\t-h  \tprint this usage information,
\t-v n\tprinting debugging information (n=1-4),
\t-d <file.h> \tprint out header for for function declaration,
\t-t  \tturn on the trace of the Yapp parser for debugging,
\t-c  \ttest the correctness of a DC file (no processing),
\t-p  \tprint out BLIFMV file,
\t<dc_file.dc>  the input specification in DC.\n";
while(scalar @ARGV>0){
   my $arg=shift @ARGV;
   if ($arg eq "-v") { $verbose=shift @ARGV; }
   elsif ($arg eq "-h") { print $Usage; exit; }
   elsif ($arg eq "-c") { $test_dc=1; }
   elsif ($arg eq "-t") { $trace_on=1; }
   elsif ($arg eq "-d") { $header_file=shift @ARGV; $print_head=1; }
   elsif ($arg eq "-p") { $print_blifmv=1; }
   else { $file=$arg; last; }
}

$file =~ /^[\w\-\/\.]+.dc$/ or die $Usage;
my $modelname = $file;
my $rootname = $file;
$rootname =~s/.dc//;
$modelname =~s/([\W|\w]+\/)*(\w+).dc/$2/g;
open (INFILE, "<$file") or die "$0: could not open file: $file\n";

my $data = join('', <INFILE>); 
$data =~ s/[\cM]+//go;
$data =~ s/(--.*\n)+//go;
my $P = new DCParse ;
print $P->version(),"\n";
$P->initialize($trace_on,$data);
my $value = $P->YYParse (yylex => \&DCParse::Lexer,
			      yyerror => \&DCParse::Error)
#                             yydebug => 0x1F )
or die "Syntax error!\n";
$P->activate();
if ($test_dc) {
    print "Read MODULE: \"",$P->module(),"\" OK!\n";
    exit;
}


#--------------------------------------------#
# The real tasks
#--------------------------------------------#
if  ($print_blifmv) {
    print_blifmv($P);
}
if  ($print_head) { 
    print_header($P,$header_file); 
}
exit;


#--------------------------------------------------------------------#
# print out in blifmv format
#--------------------------------------------------------------------#
sub print_blifmv($) {
    my ($M) = @_;
    my ($name,$cell,$net,$fi,$fo);
    my ($a,$b,$c);
    my (%{data_nodes});
    
    print "#Blifmv file automatically generated by dc2mv (".localtime().")\n";
    print ".model ",$M->module(),"\n";
    
    ##primary inputs
    print ".inputs";
    foreach $net ($M->pi()) {
	print " ",$$net{NAME};
    }
    ##primary inputs
    print "\n.outputs";
    foreach $net ($M->po()) {
	print " ",$$net{NAME};
    }
    print "\n";
    
    ##data variable declaration
    foreach $net ($M->nets()) {
	if (!($$net{TYPE}=~/Boolean/))  {
	    print ".n ",$$net{NAME},"\n" if ($$net{NAME});
	    $data_nodes{$$net{NAME}} = 1;
	}
    }
    
    ##flow definitions both control and data
    for my $i(0..scalar($M->nets())) {
	next if $i eq scalar($M->nets());
	my $equ = $M->get_equ($i);
	if ($equ) {
	    $net = $M->get_net($i);
	    if ($$net{TYPE}=~/Boolean/) {
		print_bool_func($M,$$net{NAME}, $$equ{EXPR});
	    } else {
		print_data_func($M,$$net{NAME}, $$equ{EXPR});
	    }
	}
    }
    
    ##latch definitions
    print "\n\n";
    for my $i(0..scalar($M->nets())) {
	next if $i eq scalar($M->nets());
	my $equ = $M->get_mem($i);
	next if !$equ;
	
	$net = $M->get_net($i);
	print ".reset ", $$net{NAME}, "\n";
	if ($$net{DEFAULT}) {
	    print_single_expression ($M, $$net{DEFAULT});
	} else {
	    print "0";  ##latches always default to 0
	}
	print "\n.latch ";
	print "#WARNING: latch assignment not always activated!\n" if $$equ{COND} ne "1";
	print_single_expression ($M, $$equ{EXPR});
	print " ".$$net{NAME}."\n";
    }
    print "\n.end\n";
}

#--------------------------------------------------------------------#
# print out a boolean function
#--------------------------------------------------------------------#
# inputs: (net name, expression of the net)
# -- expressions are printed recursively as separate nodes
# -- new node name is old one postfixed with "Z"
# -- call print_single_expression to print leave expressions
# -- call print_sop_table to print the logic function, const folding
#--------------------------------------------------------------------#
sub print_bool_func ($$$) {
    my ($M, $name, $exp) = (@_);
    my (@subnets, @subexps, @inputs);
    my ($a,$b);
    
    if ($$exp{TYPE} eq "CONST") {      ##easy case for constant
	print_bool_const ($M, $name, $exp); return;
    } 
    if ($$exp{TYPE} eq "FLOW") {       ##easy case for buffer
	print_bool_buffer ($M, $name, $exp); return;
    }
    ##harder case for functions
    print "#WARNING: non-defined type\n" if $$exp{TYPE} ne "FUNC"; 
    
    my $func = $$exp{FUNC};
    my $list = $$func{LIST};
    print "#WARNING: no input to boolean function\n" if ($list eq undef);
    print "#WARNING: EQ/NEQ type for boolean function\n"
	if ($$func{NAME} eq "EQ" or $$func{NAME} eq "NE");
    
    
    ##special cases for constant propagation
    my $sig_str;
    $sig_str .= $$func{NAME};
    foreach $a(@{$list}) {
	$sig_str .= "0" if ($$a{TYPE} eq "CONST" and $$a{CONST} == 1);
	$sig_str .= "1" if ($$a{TYPE} eq "CONST" and $$a{CONST} == 0);
	$sig_str .= "*" if ($$a{TYPE} eq "FLOW" or $$a{TYPE} eq "FUNC");
    }
    if ($sig_str eq "COND0**") {
	print "##condition(0, *, *)\n";
	print_bool_buffer_recur ($M, $name, @{$list}[2])
    }
    elsif ($sig_str eq "COND1**") {
	print "##condition(1, *, *)\n";
	print_bool_buffer_recur ($M, $name, @{$list}[1]);
    }
    elsif ($sig_str eq "COND*00") {
	print "##condition(*, 0, 0)\n";
	print_bool_const($M, $name, {TYPE=>"CONST",CONST=>"0"});
    }
    elsif ($sig_str eq "COND*11") {
	print "##condition(*, 1, 1)\n";
	print_bool_const($M, $name, {TYPE=>"CONST",CONST=>"1"});
    }
    elsif ($sig_str eq "COND*10") {
	print "##condition(*, 1, 0)\n";
	print_bool_buffer_recur ($M, $name, @{$list}[0]);
    }
    elsif ($sig_str eq "COND*01") {
	print "##condition(*, 0, 1)\n";
	print_bool_not_gate ($M, $name, @{$list}[0]);
    }
    elsif ($sig_str eq "AND*0" or $sig_str eq "AND0*") {
	print_bool_const($M, $name, {TYPE=>"CONST",CONST=>"0"});
    }
    elsif ($sig_str eq "AND1*") {
	print_bool_buffer_recur ($M, $name, @{$list}[1]);
    }
    elsif ($sig_str eq "AND*1") {
	print_bool_buffer_recur ($M, $name, @{$list}[0]);
    }
    elsif ($sig_str eq "OR*0") {
	print_bool_buffer_recur ($M, $name, @{$list}[0]);
    }
    elsif ($sig_str eq "OR0*") {
	print_bool_buffer_recur ($M, $name, @{$list}[1]);
    }
    elsif ($sig_str eq "OR1*" or $sig_str eq "OR*1") {
	print_bool_const($M, $name, {TYPE=>"CONST",CONST=>"1"});
    }
    elsif ($sig_str eq "NOT0") {
	print_bool_const($M, $name, {TYPE=>"CONST",CONST=>"1"});
    }
    elsif ($sig_str eq "NOT1") {
	print_bool_const($M, $name, {TYPE=>"CONST",CONST=>"0"});
    }
    elsif ($sig_str eq "XOR*0") {
	print_bool_buffer_recur ($M, $name, @{$list}[0]);
    }
    elsif ($sig_str eq "XOR0*") {
	print_bool_buffer_recur ($M, $name, @{$list}[1]);
    }
    elsif ($sig_str eq "XOR*1") {
	print_bool_not_gate ($M, $name, @{$list}[0]);
    }
    elsif ($sig_str eq "XOR1*") {
	print_bool_not_gate ($M, $name, @{$list}[1]);
    }
    else {
	##no more constants
	print_bool_sop_table ($M, $name, $func);
    }
    return;
}


#--------------------------------------------------------------------#
# print out a boolean buffer node
#--------------------------------------------------------------------#
sub print_bool_buffer_recur ($$$) {
    my ($M, $name, $exp) = (@_);
    if ($$exp{TYPE} eq "CONST") {      ##easy case for constant
	print_bool_const ($M, $name, $exp);
	return;
    }
    if ($$exp{TYPE} eq "FLOW") {       ##easy case for buffer
	print_bool_buffer ($M, $name, $exp);
	return;
    }
    my $newname = $name."Z";
    print ".table ";
    print " ".$newname." -> ".$name;
    print "0 0\n1 1\n";
    print_bool_func ($M, $newname, $exp);
    return;
}


#--------------------------------------------------------------------#
# print out a boolean constant (errot if not constant)
#--------------------------------------------------------------------#
sub print_bool_const ($$$) {
    my ($M, $name, $exp) = (@_);
    print ".table ";
    print "-> ".$name."\n";
    print_single_expression ($M, $exp);
    print "\n";
    return;
}

#--------------------------------------------------------------------#
# print out a boolean buffer (errot if not buffer/flow)
#--------------------------------------------------------------------#
sub print_bool_buffer ($$$) {
    my ($M, $name, $exp) = (@_);
    print ".table ";
    print_single_expression ($M, $exp);
    print " -> ".$name."\n";
    print "0 0\n1 1\n";
    return;
}


#--------------------------------------------------------------------#
# print out a sum-of-product table
#--------------------------------------------------------------------#
# inputs (module ref, output name, function ref)
#--------------------------------------------------------------------#
sub print_bool_sop_table ($$$) {
    my ($M, $name, $func) = @_;
    my ($a,$b);
    my $list = $$func{LIST};
    
    ##unary function
    if ($$func{NAME} eq "NOT") {
	print_bool_not_gate ($M, $name, @{$list}[0]);
	return;
    }
    elsif ($$func{NAME} eq "AND") {
	print_bool_and_gate ($M, $name, @{$list}[0], @{$list}[1]);
    }
    elsif ($$func{NAME} eq "OR") {
	print_bool_or_gate ($M, $name, @{$list}[0], @{$list}[1]);
    }
    elsif ($$func{NAME} eq "XOR") {
	print_bool_xor_gate ($M, $name, $list); ##possibly >2 inputs
    }
    elsif ($$func{NAME} eq "COND") {
	print_bool_cond_gate ($M, $name, @{$list}[0], @{$list}[1], @{$list}[2]);
    }
    else {
    }
    
    ##real boolean functions now
}

#--------------------------------------------------------------------#
# print out a not gate (no need to worry about constants)
#--------------------------------------------------------------------#
sub print_bool_not_gate ($$$) {
    my ($M, $name, $exp) = (@_);
    print ".table ";
    if ($$exp{TYPE} eq "FLOW") {
	print_single_expression ($M, $exp);
	print " -> ".$name."\n";
	print "0 1\n1 0\n"; return;
    }
    ##need recursion
    my $newname = $name."Z";
    print $newname." -> ".$name."\n";
    print "0 1\n1 0\n";
    print_func_node ($M, $newname, $exp);
    return;
}

#--------------------------------------------------------------------#
# print out an AND gate (no need to worry about constants)
#--------------------------------------------------------------------#
sub print_bool_and_gate ($$$$) {
    my ($M, $name, $op1, $op2) = (@_);
    my ($a, $tmpname, @newname, @newflag);
    my $c=0;
    $tmpname = $name;
    print ".table ";
    foreach $a($op1,$op2) {
	if ($$a{TYPE} eq "FLOW") {
	    print " ";
	    print_single_expression ($M, $a);
	    push @newname, "0";
	    push @newflag, 0;
	} else {
	    $tmpname .= "Z".$c; $c++;
	    print " ".$tmpname;
	    push @newname, $tmpname;
	    push @newflag, 1;
	}
    }
    print " -> ".$name."\n";
    print ".def 0\n1 1  1\n";
    
    ##need recursion
    print_func_node ($M, $newname[0], $op1) if $newflag[0];
    print_func_node ($M, $newname[1], $op2) if $newflag[1];
    return;
}

#--------------------------------------------------------------------#
# print out an OR gate (no need to worry about constants)
#--------------------------------------------------------------------#
sub print_bool_or_gate ($$$$) {
    my ($M, $name, $op1, $op2) = (@_);
    my ($a, $tmpname, @newname, @newflag);
    my $c=0;
    $tmpname = $name;
    print ".table ";
    foreach $a($op1,$op2) {
	if ($$a{TYPE} eq "FLOW") {
	    print " ";
	    print_single_expression ($M, $a);
	    push @newname, "0";
	    push @newflag, 0;
	} else {
	    $tmpname .= "Z".$c; $c++;
	    print " ".$tmpname;
	    push @newname, $tmpname;
	    push @newflag, 1;
	}
    }
    print " -> ".$name."\n";
    print ".def 1\n0 0  0\n";
    
    ##need recursion
    print_func_node ($M, $newname[0], $op1) if $newflag[0];
    print_func_node ($M, $newname[1], $op2) if $newflag[1];
    return;
}

#--------------------------------------------------------------------#
# print out an XOR gate (no need to worry about constants)
#--------------------------------------------------------------------#
#--------------------------------------------------------------------#
sub print_bool_xor_gate ($$$) {
    my ($M, $name, $list) = (@_);
    my ($a, $tmpname, $size, @newname, @newflag);
    my @xor = (".def 0\n0 1 1\n1 0 1\n",
	   ".def 0\n1 0 0 1\n0 1 0 1\n0 0 1 1\n1 1 1 1\n",
	   ".def 0\n1 0 0 0 1\n0 1 0 0 1\n0 0 1 0 1\n1 1 1 0 1\n0 1 1 1 1\n 1 0 1 1 1\n 1 1 0 1 1\n 1 1 1 0 1\n",);
    my $c=0;
    $tmpname = $name;
    $size    = scalar(@{$list}) - 1;
    
    print ".table ";
    foreach $a(@{$list}) {
	if ($$a{TYPE} eq "CONST") {
	    print "WARNING: xor (0/1, *,...)\n";
	    next;
	}
	if ($$a{TYPE} eq "FLOW") {
	    print " ";
	    print_single_expression ($M, $a);
	    push @newname, "0";
	    push @newflag, 0;
	} else {
	    $tmpname .= "Z".$c; $c++;
	    print " ".$tmpname;
	    push @newname, $tmpname;
	    push @newflag, 1;
	}
    }
    print " -> ".$name."\n";
    print xor[$size-2];
    
    foreach $a(@{$list}) {
	print_func_node ($M, $newname[0], $a) if $newflag[0];
    }
    return;
}

#--------------------------------------------------------------------#
# print out a boolean condition (*,*,*) node
#--------------------------------------------------------------------#
# -- consider cases for (*, 0/1, 0/1) constants (can't be both const!)
#--------------------------------------------------------------------#
sub print_bool_cond_gate ($$$$$) {
    my ($M, $name, $op1, $op2, $op3) = (@_);
    my %cond_tbl = ("cond***"=>"1 0 - 0\n1 1 - 1\n0 - 0 0\n0 - 1 1\n",
		    "cond*0*"=>"1 - 0\n0 0 0\n0 1 1\n",
		    "cond**0"=>"1 0 0\n1 1 1\n0 - 0\n",
		    "cond*1*"=>"1 - 1\n0 0 0\n0 1 1\n",
		    "cond**1"=>"1 0 0\n1 1 1\n0 - 1\n");
    
    my ($a, $tmpname, $funname, @newname, @newflag);
    my $c=0;
    $tmpname = $name;
    $funname = "cond";
    print ".table ";
    foreach $a($op1,$op2,$op3) {
	if ($$a{TYPE} eq "CONST") {
	    $funname .= "0" if ($$a{CONST} eq "0");
	    $funname .= "1" if ($$a{CONST} eq "1");
	    push @newname, "0";
	    push @newflag, 0;
	} elsif ($$a{TYPE} eq "FLOW") {
	    $funname .= "*";
	    print " ";
	    print_single_expression ($M, $a);
	    push @newname, "0";
	    push @newflag, 0;
	} else {
	    $funname .= "*";
	    $tmpname .= "Z".$c; $c++;
	    print " ".$tmpname;
	    push @newname, $tmpname;
	    push @newflag, 1;
	}
    }
    print " -> ".$name."\n";
    print $cond_tbl{$funname};
    
    ##need recursion
    print_func_node ($M, $newname[0], $op1) if $newflag[0];
    print_func_node ($M, $newname[1], $op2) if $newflag[1];
    print_func_node ($M, $newname[2], $op3) if $newflag[2];
    return;
}


######################################################################
######################################################################
#--------------------------------------------------------------------#
# print out a data node (mux or expr)
#--------------------------------------------------------------------#
# -- for nested functions, traverse to leaves for true support
# -- for constants (#0,#1) in mux, create new constant nodes (hash!)
# -- can we do buffer removal??
#--------------------------------------------------------------------#
my %const_nodes;
sub print_data_func ($$$) {
    my ($M, $name, $exp) = (@_);
    
    if ($$exp{TYPE} eq "CONST") {      ##easy case for constant
	print ".data -> ".$name."\n\"";
	print_single_expression ($M,$exp); 
	print "\"\n";
	return;
    } 
    if ($$exp{TYPE} eq "FLOW") {       ##easy case for buffer
	print ".data ";
	print_single_expression ($M,$exp);
	print "-> ".$name."\n\"";
	print_single_expression ($M,$exp);
	print "\"\n";
	return;
    }
    ##harder case for functions
    print "#WARNING: non-defined type\n" if $$exp{TYPE} ne "FUNC"; 
    my $func = $$exp{FUNC};
    my $list = $$func{LIST};
    
    if ($list eq undef) {               ##void function calls
	print ".data -> ".$name."\n\"";
	print_single_expression ($M,$exp); 
	print "\"\n";
	return;
    }
    
    ##need to traverse to the expression leaves to get true support!!
    if ($$func{NAME} eq "COND") {       ##mux gate
	print_data_mux ($M, $name, $func, $list);
	return;
    }
    print ".data ";
    
    my @param_list = util_data_param_list_recur($list);
    foreach my $a(@param_list) {
	next if ($$a{TYPE} eq "CONST");
	print " ";
	print_single_expression ($M,$a);
    }
    print " -> ".$name."\n";
    print "\"";
    print_data_expression_recur($M,$name,$exp);
    print "\"\n";
}

#--------------------------------------------------------------------#
# print out a data expression recursively (assume type FUNC)
#--------------------------------------------------------------------#
#--------------------------------------------------------------------#
sub print_data_expression_recur ($$$) {
    my ($M,$name,$exp) = (@_);
    my %func_sym = ("COND"=>"\?", "EQ"=>"==", "NEQ"=>"\!=", "OR"=>"\|",
     "AND"=>"\&",  "NOT"=>"\!",  "XOR"=>"\^",
     "IEQ"=>"==", "INEQ"=>"\!=", "ILT"=>"\<", "ILEQ"=>"\<=", "IGT"=>"\>", "IGEQ"=>"\>=", 
     "IPLUS"=>"\+","IMINUS"=>"-","IMUL"=>"*","IMOD"=>"\%",  "IDIV"=>"\/","IUMINUS"=>"--",
     "FEQ"=>"==", "FNEQ"=>"\!=", "FLT"=>"\<", "FLEQ"=>"\<=", "FGT"=>"\>", "FGEQ"=>"\>=", 
     "FPLUS"=>"\+","FMINUS"=>"-","FMUL"=>"*","FDIV"=>"\/",  "FUMINUS"=>"--",
     "DEQ"=>"==", "DNEQ"=>"\!=", "DLT"=>"\<", "DLEQ"=>"\<=", "DGT"=>"\>", "DGEQ"=>"\>=", 
     "DPLUS"=>"\+","DMINUS"=>"-","DMUL"=>"*","DDIV"=>"\/",  "DUMINUS"=>"--",);
    
    if ($$exp{TYPE} eq "CONST" or $$exp{TYPE} eq "FLOW") {
	print_single_expression ($M,$exp); 
	return;
    } 
    
    my $func = $$exp{FUNC};
    my $list = $$func{LIST};
    
    if ($func_sym{$$func{NAME}} ne undef) {
	if (scalar(@{$list}) == 1) {        ##unary function calls
	    print "(";
	    print_data_expression_recur ($M,$name,@{$list}[0]);
	    print ")";
	    print "--" if ($$func{NAME} =~ /UMINUS/);
	    return;
	}
	#print "(";
	print_data_expression_recur ($M,$name,@{$list}[0]);
	print " ".$func_sym{$$func{NAME}}." ";
	print_data_expression_recur ($M,$name,@{$list}[1]);
	#print ")";
    }
    else {                ##external or imported functions
	my $func_pt = $$func{NAME};
	my $size    = scalar(@{$list}) - 1;
	print $$func_pt{NAME}."(";
	print_data_expression_recur ($M,$name,@{$list}[0]);
	foreach my $a(1..$size) {
	    print ", ";
	    print_data_expression_recur ($M,$name,@{$list}[$a]);
	}
	print ")";
    }
}


#--------------------------------------------------------------------#
# print out a mux expression (assume no constants)
#--------------------------------------------------------------------#
# -- create new nodes for constants and print recursively
#--------------------------------------------------------------------#
sub print_data_mux ($$$$) {
    my ($M, $name, $fun, $list) = (@_);
    my (@newname);
    my $tmpname = $name;
    
    my $op0 = @{$list}[0];
    my $op1 = @{$list}[1];
    my $op2 = @{$list}[2];
    my $c=0;
    
    ##control signal is a logic function
    if ($$op0{TYPE} eq "FUNC") {
	$tmpname .= "Z".$c; $c++;
	print_bool_func ($M, $tmpname, $op0);
	push @newname, $tmpname;
    } else {
	push @newname, get_expression_str($M,$op0);
    }
    foreach my $a($op1, $op2) {
	if ($$a{TYPE} eq "FUNC") {
	    $tmpname .= "Z".$c; $c++;
	    push @newname, $tmpname;
	    print ".n ".$tmpname."\n";
	    print_data_func ($M, $tmpname, $a);
	}
	elsif ($$a{TYPE} eq "CONST") {
	    if ($const_nodes{$$a{CONST}}) {
		push @newname, $const_nodes{$$a{CONST}};
	    } else {
		$tmpname .= "Z".$c; $c++;
		$const_nodes{$$a{CONST}} = $tmpname;
		push @newname, $tmpname;
		print ".n ".$tmpname."\n";
		print_data_func ($M, $tmpname, $a);
	    }
	}
	else {
	    push @newname, get_expression_str($M,$a);
	}
    }
    print ".mux ";
    foreach my $a(@newname) {
	print " ".$a;
    }
    print "-> ".$name."\n";
    print "0 - -  =".$newname[2]."\n";
    print "1 - -  =".$newname[1]."\n";
}


#--------------------------------------------------------------------#
# print out a leaf expression (no recursion)
#--------------------------------------------------------------------#
sub print_single_expression ($$) {
    my ($M, $expr) = (@_);
    my ($const);
    my %func_sym = ("COND"=>"\?", "EQ"=>"==", "NEQ"=>"\!=", "OR"=>"\|",
     "AND"=>"\&",  "NOT"=>"\!",  "XOR"=>"\^",
     "IEQ"=>"==", "INEQ"=>"\!=", "ILT"=>"\<", "ILEQ"=>"\<=", "IGT"=>"\>", "IGEQ"=>"\>=", 
     "IPLUS"=>"\+","IMINUS"=>"-","IMUL"=>"*","IMOD"=>"\%",  "IDIV"=>"\/","IUMINUS"=>"--",
     "FEQ"=>"==", "FNEQ"=>"\!=", "FLT"=>"\<", "FLEQ"=>"\<=", "FGT"=>"\>", "FGEQ"=>"\>=", 
     "FPLUS"=>"\+","FMINUS"=>"-","FMUL"=>"*","FDIV"=>"\/",  "FUMINUS"=>"--",
     "DEQ"=>"==", "DNEQ"=>"\!=", "DLT"=>"\<", "DLEQ"=>"\<=", "DGT"=>"\>", "DGEQ"=>"\>=", 
     "DPLUS"=>"\+","DMINUS"=>"-","DMUL"=>"*","DDIV"=>"\/",  "DUMINUS"=>"--",);

    if ($$expr{TYPE} eq "CONST") {
	$const = $$expr{CONST};
	if ($const =~ /^\d+([E|D])*(\d+)*/) {
	    print $const;
	} else {
	    print $$const{NAME};
	}
    }
    elsif ($$expr{TYPE} eq "FLOW") {
	my $net = $M->get_net($$expr{FLOW});
	print $$net{NAME};
    }
    elsif ($$expr{TYPE} eq "FUNC") {
	my $func = $$expr{FUNC};
	if ($func_sym{$$func{NAME}} ne undef) {
	    print $$func{NAME}." ()";
	} else {
	    my $func_pt = $$func{NAME};
	    print $$func_pt{NAME}." ()";
	}
    }
    else {
    }
}

#--------------------------------------------------------------------#
# return a leaf expression as string (no recursion)
#--------------------------------------------------------------------#
sub get_expression_str($$) {
    my ($M, $expr) = (@_);
    if ($$expr{TYPE} eq "CONST") {
	return $$expr{CONST};
    }
    elsif ($$expr{TYPE} eq "FLOW") {
	my $net = $M->get_net($$expr{FLOW});
	return $$net{NAME};
    }
    elsif ($$expr{TYPE} eq "FUNC") {
	my $func = $$expr{FUNC};
	return $$func{NAME}." ()";
    }
    else {
    }
}

#--------------------------------------------------------------------#
# Print out a fuction node recursively
#--------------------------------------------------------------------#
sub print_func_node ($$$) {
    my ($M,$name,$exp) = (@_);
    if (test_exp_is_ctr ($M, $exp)) {
	print_bool_func ($M, $name, $exp);
    } else {
	print_data_func ($M, $name, $exp);
    }
}

#--------------------------------------------------------------------#
# Test if an expression is control or data
#--------------------------------------------------------------------#
# -- Assume the expression is either a flow or function
#--------------------------------------------------------------------#
sub test_exp_is_ctr ($$) {
    my ($M, $exp) = (@_);
    
    if ($$exp{TYPE} eq "FLOW") {
	my $net = $M->get_net($$exp{FLOW});
	return ($$net{TYPE} =~ /Boolean/);
    }
    return 0 if ($$exp{TYPE} ne "FUNC");
    
    my $func = $$exp{FUNC};
    my $list = $$func{LIST};
    
    foreach my $a(@{$list}) {
	return 1 if (test_exp_is_ctr($M, $a));
    }
    return 0;
}

#--------------------------------------------------------------------#
# Test if an expression is constant or not
#--------------------------------------------------------------------#
sub test_exp_is_const($) {
    return 1 if $$_{TYPE} eq "CONST";
    return 0;
}

#--------------------------------------------------------------------#
# Return the leaves of the given parameter tree
#--------------------------------------------------------------------#
sub util_data_param_list_recur($) {
    my ($list) = (@_);
    my (@params, $a);
    my ($a, $b, $func, $inlist, @inparams);
    
    foreach $a(@{$list}) {
	next if $$a{TYPE} eq "CONST";
	if ($$a{TYPE} eq "FLOW") {
	    push (@params, $a) ;
	}
	elsif ($$a{TYPE} eq "FUNC") {
	    #print "nested function";
	    $func = $$a{FUNC};
	    $inlist = $$func{LIST};
	    @inparams = util_data_param_list_recur($inlist);
	    foreach $b(@inparams) { push @params, $b; }
	}
    }
    return @params;
}


######################################################################
######################################################################
#--------------------------------------------------------------------#
# print out header file for procedures/functions/constants
#--------------------------------------------------------------------#
sub print_header ($$) {
    my ($M,$fp) = (@_);
    my ($a,$b);
    
    open (HFILE, ">$fp") or die "could not open file: $fp\n";
    print HFILE "/* [".$M->module()."] header file generated by dc2mv (".localtime().") */\n\n";

    print HFILE "/* constants */\n";
    foreach $a ($M->const()) {
	if ($$a{EXT} eq "1") {
	    print HFILE "extern ".$$a{NAME}.";\n";
	} else {
	    print HFILE "#define ".$$a{NAME}." ".$$a{VALUE}."\n";
	}
    }
    print HFILE "\n/* procedures */\n";
    foreach $a ($M->proc()) {
	print HFILE "extern ".$$a{NAME}."();\n"
    }
    print HFILE "\n/* functions */\n";
    foreach $a ($M->func()) {
	print HFILE "extern ".$$a{NAME}."();\n"
    }
    close (HFILE);
    return;
}
